import 'dart:ffi';
import 'dart:typed_data';
import 'package:ffi/ffi.dart';
import 'array_utils.dart';

import 'array.dart';

/// Instead of doing many little allocations / deallocations, Arena
/// allocates / deallocates a large chunk of memory for multiple pointers / arrays,
/// which can improve performance and makes memory management simpler.
///
/// Note: Don't free slices of Arena manually. Only call [delete] after you are
/// done with all slices. Using slices after deleting Arena leads to undefined behaviour.
/// Not deleting Arena will cause a memory leak.
/// ````
/// final arena = Arena<Uint8,Uint8List>(10, (ptr, length) => ptr.asTypedList(length));
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.view.length, array2.ptr, array2.view.length);
/// arena.delete();
/// ````
class Arena<T extends NativeType, L extends List<int>> {
  int _index = 0;

  /// Total length of allocated memory
  final int length;
  final Pointer<T> _arrPtr;
  final L Function(Pointer<T>, int) _getView;

  /// Left free space of Arena.
  int get freeSpace => length - _index;

  /// [length] of allocated memory.
  /// [_getView] returns the associated typed list for [Pointer<T>].
  Arena(this.length, this._getView) : _arrPtr = allocate(count: length);

  /// [_arrPtr] points to a pre-allocated array of [length].
  Arena.fromPtr(this._arrPtr, this.length, this._getView);

  /// Returns a slice of [length] from the pre-allocated array of Arena.
  /// When [length] exceeds [freeSpace] it throws a [RangeError].
  Array<T, L> newArray(int length) {
    assert(length > 0);
    if (freeSpace < length) {
      throw RangeError('Not enough free space');
    }
    final addr = _arrPtr.address + _index;
    final newPtr = Pointer<T>.fromAddress(addr);
    final view = _getView(newPtr, length);
    _index += length;
    return Array<T, L>.fromPtr(newPtr, view);
  }

  /// Deletes all allocated memory.
  void delete() {
    free(_arrPtr);
  }
}

class Uint8Arena extends Arena<Uint8, Uint8List> {
  Uint8Arena(int length) : super(length, getUint8View);
  Uint8Arena.fromPtr(Pointer<Uint8> ptr, int length)
      : super.fromPtr(ptr, length, getUint8View);
}
