import 'dart:typed_data';

import 'package:ffi_helper/src/array_utils.dart';

import 'array.dart';
import 'dart:ffi';
import 'package:ffi/ffi.dart';

/// Instead of allocating / deallocating many arrays, share one array which
/// gets deleted when all references to it got deleted. This can lead to
/// better performance and reduced memory usage.
///
/// Gets lazily initialized when the first reference gets created.
/// It is safe to use even after deletion; it just reallocates new memory for new references.
/// Not deleting all references causes memory leaks.
/// ````
/// final shared = SharedArray<Uint8,Uint8List>(10, (ptr,len) => ptr.asTypedList(len));
/// // All instances of [WritesToArray] share the same [SharedArray] so they don't have to all
/// // allocate their own array.
/// class WritesToArray {
///   final _buffer = shared.newRef();
///
///   Uint8List someNativeWrapper(){
///     writeToBuffer(_buffer.ptr);
///     return Uint8List.fromList(_buffer.view);
///   }
///   void delete() => _buffer.delete();
/// }
/// ````
class SharedArray<T extends NativeType, L extends List<int>> {
  int _refCount = 0;
  int get refCount => _refCount;
  Array<T, L> _array;
  final L Function(Pointer<T>, int) _getView;
  final int length;

  SharedArray(this.length, this._getView);

  /// Create new reference to this SharedArray.
  ArrayRef<T, L> newRef() {
    if (_refCount == 0) {
      _array = Array(length, _getView);
    }
    _refCount++;
    return ArrayRef._(this);
  }

  void _delRef() {
    assert(_refCount > 0);
    _refCount--;
    if (_refCount == 0) {
      free(_array.ptr);
    }
  }
}

/// Reference to a [SharedArray].
class ArrayRef<T extends NativeType, L extends List<int>>
    implements Array<T, L> {
  final SharedArray<T, L> _shared;
  @override
  Pointer<T> get ptr => _shared._array.ptr;
  @override
  L get view => _shared._array.view;
  bool _freed = false;

  ArrayRef._(this._shared);
  @override
  void delete() {
    if (_freed) {
      return;
    }
    _shared._delRef();
    _freed = true;
  }
}

class SharedUint8 extends SharedArray<Uint8, Uint8List> {
  SharedUint8(int length) : super(length, getUint8View);
}

class Uint8Ref extends ArrayRef<Uint8, Uint8List> {
  Uint8Ref._(SharedUint8 shared) : super._(shared);
}
